"""DEPRECATED.

This module has been moved to ocflib.printing.printers. In a future version,
this module will be removed entirely.
"""
from ocflib.printing.printers import _snmp  # noqa
from ocflib.printing.printers import get_lifetime_pages  # noqa
from ocflib.printing.printers import get_maintkit  # noqa
from ocflib.printing.printers import get_toner  # noqa
from ocflib.printing.printers import OID_LIFETIME_PAGES_PRINTED  # noqa
from ocflib.printing.printers import OID_MAINTKIT_CUR  # noqa
from ocflib.printing.printers import OID_MAINTKIT_MAX  # noqa
from ocflib.printing.printers import OID_TONER_CUR  # noqa
from ocflib.printing.printers import OID_TONER_MAX  # noqa
from ocflib.printing.printers import PRINTERS  # noqa

# TODO: fix callers of this module



"""Networking."""
from ipaddress import ip_address
from ipaddress import ip_network
from ipaddress import IPv4Address
from ipaddress import IPv6Address

OCF_SUBNET_V4 = ip_network('169.229.226.0/24')
OCF_SUBNET_V6 = ip_network('2607:f140:8801::/64')
OCF_SUBNET_V6_COMPAT = ip_network('2607:f140:8801::1:0/112')
OCF_DNS_RESOLVER = ip_address('169.229.226.22')


def ipv6_to_ipv4(ipv6):
    """Convert an OCF IPv6 address to its equivalent IPv4.

    This works only for IPs in our "compatibility" subnet, which is a special
    subnet where we map one-to-one hosts from IPv4 to IPv6.

    Note that the bits do *not* map exactly, since IPv6 is written in
    hexadecimal, whereas IPv4 is written in decimal. For the benefit of humans,
    we make the human-readable portion the same, not the binary representation.
    """
    assert isinstance(ipv6, IPv6Address)
    assert ipv6 in OCF_SUBNET_V6_COMPAT
    last_group = int(ipv6.exploded.split(':')[-1])
    return ip_address(
        int.from_bytes(OCF_SUBNET_V4.network_address.packed, 'big') | last_group
    )


def ipv4_to_ipv6(ipv4):
    """Convert an OCF IPv4 address to its equivalent compatibility IPv4.

    For a description of the compability IPv6 subnet, see `ipv6_to_ipv4` above.
    """
    assert isinstance(ipv4, IPv4Address)
    assert ipv4 in OCF_SUBNET_V4
    last_group = int(ipv4.exploded.split('.')[-1], 16)
    return ip_address(
        int.from_bytes(OCF_SUBNET_V6_COMPAT.network_address.packed, 'big') | last_group
    )


def is_ocf_ip(ip):
    """Return whether this IP is owned by the OCF.

    Accepts both IPv4 and IPv6 addresses in object form from the ipaddress
    module.

    >>> from ipaddress import ip_address
    >>> is_ocf_ip(ip_address('169.229.226.12'))
    True
    """
    if isinstance(ip, IPv4Address):
        return ip in OCF_SUBNET_V4
    elif isinstance(ip, IPv6Address):
        return ip in OCF_SUBNET_V6
    else:
        raise AssertionError('You must pass in an IPv4Address or IPv6Address object.')



from base64 import b64encode

import mock
import pytest

from ocflib.infra.ldap import create_ldap_entry_with_keytab


@pytest.yield_fixture
def mock_spawn():
    with mock.patch('pexpect.spawn') as spawn:
        yield spawn


@pytest.yield_fixture
def mock_send_problem_report():
    with mock.patch('ocflib.infra.ldap.send_problem_report') as m:
        yield m


class TestCreateLdapEntry:

    def test_normal_creation(self, mock_spawn, mock_send_problem_report):
        mock_spawn.return_value.before = b'\n'
        create_ldap_entry_with_keytab(
            'uid=ckuehl,ou=People,dc=OCF,dc=Berkeley,dc=EDU',
            {'a': ['b', 'c'], 'd': ['e']},
            '/nonexist',
            'create/admin',
        )

        mock_spawn.assert_called_with(
            'kinit -t /nonexist create/admin ldapadd',
            timeout=10,
        )
        mock_spawn.return_value.expect.assert_has_calls([
            mock.call('SASL data security layer installed.'),
            mock.call('adding new entry "uid=ckuehl,ou=People,dc=OCF,dc=Berkeley,dc=EDU"'),
        ])

        def encode(attr, value):
            return '{attr}:: {value}'.format(
                attr=attr,
                value=b64encode(value.encode('utf8')).decode('ascii'),
            )

        mock_spawn.return_value.sendline.assert_has_calls([
            mock.call(encode(attr, value))
            for attr, value in [
                ('dn', 'uid=ckuehl,ou=People,dc=OCF,dc=Berkeley,dc=EDU'),
                ('a', 'b'),
                ('a', 'c'),
                ('d', 'e'),
            ]
        ], any_order=True)
        assert mock_spawn.return_value.sendeof.called
        assert not mock_send_problem_report.called

    def test_already_exists(self, mock_spawn, mock_send_problem_report):
        mock_spawn.return_value.before = b'\nAlready exists (68)\n'
        with pytest.raises(ValueError):
            create_ldap_entry_with_keytab(
                'uid=ckuehl,ou=People,dc=OCF,dc=Berkeley,dc=EDU',
                {'a': ['b', 'c'], 'd': ['e']},
                '/nonexist',
                'create/admin',
            )
        assert not mock_send_problem_report.called

    def test_unexpected_error(self, mock_spawn, mock_send_problem_report):
        mock_spawn.return_value.before = b'\nlol wut is this error\n'
        with pytest.raises(ValueError):
            create_ldap_entry_with_keytab(
                'uid=ckuehl,ou=People,dc=OCF,dc=Berkeley,dc=EDU',
                {'a': ['b', 'c'], 'd': ['e']},
                '/nonexist',
                'create/admin',
            )
        assert mock_send_problem_report.called



import pytest

from ocflib.infra.hosts import hostname_from_domain
from ocflib.infra.hosts import hosts_by_filter
from ocflib.infra.hosts import type_of_host


class TestHostsByFilter:

    def _hostnames(self, results):
        return [entry['cn'][0] for entry in results]

    @pytest.mark.parametrize('filter_str,expected', [
        ('(cn=death)', ['death']),
        ('(cn=doesnotexist)', []),
        ('(herp=derp)', []),
    ])
    def test_hosts_by_filter(self, filter_str, expected):
        results = self._hostnames(hosts_by_filter(filter_str))
        assert set(results) == set(expected)

    @pytest.mark.parametrize('filter_str', ['', 'cn=death', '42', 'asdf'])
    def test_invalid_filters(self, filter_str):
        with pytest.raises(Exception):
            hosts_by_filter(filter_str)

    def test_puppet_class(self):
        # This will break if death is ever renamed, but it's a useful test.
        #
        # We choose to test death because it is in lots of university DNS
        # records, so it is probably one of the more unlikely hosts to be
        # renamed.
        assert ('death' in
                self._hostnames(hosts_by_filter('(puppetClass=ocf_www)')))


@pytest.mark.parametrize('fqdn,expected', [
    ('death.ocf.berkeley.edu', 'death'),
    ('death', 'death'),
    ('', ''),
])
def test_hostname_from_domain(fqdn, expected):
    assert hostname_from_domain(fqdn) == expected

# This will similarly break if death or eruption are renamed.
#
# The former is unlikely, as explained above. The latter is less likely than
# any other desktop, since it is configured specially as the staff-only
# computer and the hostname will probably be reused when desktop machines are
# changed out.


@pytest.mark.parametrize('hostname,expected', [
    ('death.ocf.berkeley.edu', None),
    ('death', 'server'),
    ('eruption', 'desktop'),
    ('doesnotexist', None),
    ('', None),
])
def test_type_of_host(hostname, expected):
    assert type_of_host(hostname) == expected



import os
import string

import pexpect

from ocflib.constants import KADMIN_PATH
from ocflib.misc.shell import escape_arg


def create_kerberos_principal_with_keytab(
    principal,
    keytab,
    admin_principal,
    password=None,
):
    """Creates a Kerberos principal by shelling out to kadmin.

    :param principal: name of the principal to create
    :param admin_principal: name of the admin principal
    :param keytab: path to the admin keytab
    :param password: password of the new principal (optional);
                     if not given, defaults to using a random password
    :param admin_principal: admin principal to use with the keytab
    :return: the password of the newly-created account
    """
    # try changing using kadmin pexpect
    cmd = ('{kadmin} -K {keytab} -p {admin} add --use-defaults ' +
           '{principal}').format(
        kadmin=escape_arg(KADMIN_PATH),
        keytab=escape_arg(keytab),
        admin=escape_arg(admin_principal),
        principal=escape_arg(principal),
    )

    if not password:
        # XXX: using `--random-password` generates weak passwords, plus spits
        # them to stdout, so we just generate a random one ourselves
        allowed = string.ascii_letters + string.digits + string.punctuation
        password = ''.join(allowed[byte % len(allowed)]
                           for byte in os.urandom(100))

    child = pexpect.spawn(cmd, timeout=10)

    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(principal))
    child.sendline(password)
    child.expect("Verify password - {}@OCF.BERKELEY.EDU's Password:"
                 .format(principal))
    child.sendline(password)

    child.expect(pexpect.EOF)

    output = child.before.decode('utf8')
    if 'kadmin' in output:
        raise ValueError('kadmin Error: {}'.format(output))

    return password



import mock
import pytest

from ocflib.account.manage import _notify_password_change
from ocflib.account.manage import change_password_with_keytab
from ocflib.account.manage import change_password_with_staffer


@pytest.yield_fixture
def mock_spawn():
    with mock.patch('pexpect.spawn') as spawn:
        yield spawn


@pytest.yield_fixture
def mock_notify_password_change():
    with mock.patch(
        'ocflib.account.manage._notify_password_change',
    ) as _notify_password_change:
        yield _notify_password_change


class TestChangePasswordWithStaffer:

    def _chpass(self, mock_spawn):
        change_password_with_staffer(
            'ggroup',
            'strong_hunter2',
            'ckuehl/admin',
            'super_hunter2',
        )
        mock_spawn.assert_called_with(
            '/usr/bin/kadmin -p ckuehl/admin cpw ggroup',
            timeout=10,
        )
        mock_spawn.return_value.sendline.assert_has_calls(
            [mock.call('strong_hunter2'), mock.call('strong_hunter2')],
        )

    def test_success(self, mock_spawn, mock_notify_password_change):
        self._chpass(mock_spawn)
        mock_notify_password_change.assert_called_once_with('ggroup', comment=None)

    @pytest.mark.parametrize('error', [
        'kadmin: cpw ggroup: Looping detected inside krb5_get_in_tkt',
        'kadmin: some random error you do not expect',
    ])
    def test_kadmin_failure(
        self,
        error,
        mock_spawn,
        mock_notify_password_change,
    ):
        mock_spawn.return_value.before.decode.return_value = error
        with pytest.raises(ValueError):
            self._chpass(mock_spawn)
        assert not mock_notify_password_change.called


class TestChangePasswordWithKeytab:

    def _chpass(self, mock_spawn):
        change_password_with_keytab(
            'ggroup',
            'strong_hunter2',
            '/some/keytab',
            'create/admin',
        )
        mock_spawn.assert_called_with(
            '/usr/bin/kadmin -K /some/keytab -p create/admin cpw ggroup',
            timeout=10,
        )
        mock_spawn.return_value.sendline.assert_has_calls(
            [mock.call('strong_hunter2'), mock.call('strong_hunter2')],
        )

    def test_success(self, mock_spawn, mock_notify_password_change):
        self._chpass(mock_spawn)
        mock_notify_password_change.assert_called_once_with('ggroup', comment=None)

    def test_kadmin_failure(self, mock_spawn, mock_notify_password_change):
        mock_spawn.return_value.before.decode.return_value = (
            'kadmin: cpw ggroup: Looping detected inside krb5_get_in_tkt'
        )
        with pytest.raises(ValueError):
            self._chpass(mock_spawn)
        assert not mock_notify_password_change.called


class TestNotifyPasswordChange:

    @mock.patch('ocflib.misc.mail.send_mail_user')
    def test_without_comment(self, mock_send_mail_user):
        _notify_password_change('ckuehl')
        mock_send_mail_user.assert_called_once_with('ckuehl', mock.ANY, mock.ANY)

    @mock.patch('ocflib.misc.mail.send_mail_user')
    def test_with_comment(self, mock_send_mail_user):
        _notify_password_change('ckuehl', comment='HERPDERP')
        mock_send_mail_user.assert_called_once_with('ckuehl', mock.ANY, mock.ANY)
        assert '\nHERPDERP\n' in mock_send_mail_user.call_args_list[0][0][2]



from setuptools import find_packages
from setuptools import setup

try:
    with open('.version') as f:
        VERSION = f.readline().strip()
except IOError:
    VERSION = 'unknown'

setup(
    name='ocflib',
    version=VERSION,
    author='Open Computing Facility',
    author_email='help@ocf.berkeley.edu',
    description='libraries for account and server management',
    url='https://www.ocf.berkeley.edu',
    packages=find_packages(),
    package_data={
        'ocflib.printing': ['ocfprinting.schema'],
    },
    install_requires=(
        'cached_property',
        'cracklib',
        'dnspython3',
        'ldap3',
        'paramiko',
        'pexpect',
        'pycrypto',
        'pymysql',
        'pysnmp',
        'pyyaml',
        'redis',
        'requests',
        'sqlalchemy',
    ),
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
)



import difflib
import pwd
import string
import sys

import cracklib

import ocflib.constants as constants
import ocflib.misc.mail


def validate_username(username, check_exists=False):
    """Validate a username, raising a descriptive exception if problems are
    encountered."""

    if username_reserved(username):
        raise ValueError('Username is reserved')

    if not 3 <= len(username) <= 16:
        raise ValueError('Username must be between 3 and 16 characters')

    if not all(c.islower() for c in username):
        raise ValueError('Username must be all lowercase letters')

    if check_exists and not user_exists(username):
        raise ValueError('Username does not exist')


def validate_password(username, password, strength_check=True):
    """Validate a password, raising a descriptive exception if problems are
    encountered. Optionally checks password strength."""

    if strength_check:
        if len(password) < 8:
            raise ValueError('Password must be at least 8 characters')

        s = difflib.SequenceMatcher()
        s.set_seqs(password, username)

        if s.ratio() > 0.6:
            raise ValueError('Password is too similar to username')

        try:
            cracklib.FascistCheck(password)
        except ValueError as e:
            raise ValueError('Password problem: {}'.format(e))

    # sanity check; note we don't use string.whitespace since we don't want
    # tabs or newlines
    allowed = string.digits + string.ascii_letters + string.punctuation + ' '

    if not all(c in allowed for c in password):
        raise ValueError('Password contains forbidden characters')


# TODO: we have two implementations of this (one here, one in search).
# one should be removed.
def user_exists(username):
    try:
        pwd.getpwnam(username)
    except KeyError:
        return False
    else:
        return True


def username_reserved(username):
    if username.startswith('ocf'):
        return True

    if username in constants.RESERVED_USERNAMES:
        return True

    # sanity check: make sure no local users share the username
    with open('/etc/passwd') as f:
        if any(line.startswith(username + ':') for line in f):
            print(
                'WARNING: Username {} rejected based on /etc/passwd!'
                .format(username),
                file=sys.stderr)
            ocflib.misc.mail.send_problem_report(
                """Username {} rejected based on /etc/passwd. It should be \
added to RESERVED_USERNAMES for consistency across \
servers!""".format(username))
            return True

    return False
